// Relarive Color Area Correction tool for ART Ver. J2.0
// Copyright (c) 2023-24 Ohnishi, Yasuo 

//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with ART.  If not, see <http://www.gnu.org/licenses/>.



// @ART-colorspace: "rec2020"
// @ART-label: "相対色領域補正"


float[3] rgb_in_area(varying float p_rgb[3], float adj_rgb[4], float adj_L,
            float amount, float Strength, bool bl_AreaPrView)
{
    float rgb[3];
    if (bl_AreaPrView == false){// do not show target area
        for (int i = 0; i < 3; i = i+1) {
            rgb[i] = p_rgb[i] + adj_rgb[i] * amount * 4 * Strength;
            rgb[i] = rgb[i] * adj_L;
        }
    }else{// show target area
        rgb[0] = amount * 4 * Strength;
        rgb[1] = amount * 4 * Strength;
        rgb[2] = 0.0;
    }
    return rgb;
}

float[3] rgb_out_area(varying float p_rgb[3], bool bl_AreaPrView)
{
    float rgb[3];
    if (bl_AreaPrView == false){// do not show target area
        for (int i = 0; i < 3; i = i+1) {
            rgb[i] = p_rgb[i];
        }
    }else{// show target area
        for (int i = 0; i < 3; i = i+1) {
            rgb[i] = 0.0;
        }
    }
    return rgb;
}



float gamma22(varying float v)
{
    float v2;
    v2 = pow(v, 1 / 2.2);
    return v2;
}

float[3] colorCastCr(varying float r, varying float g,varying float b,
                    float adj_rgb[4], int param_choice,
                    float Pivot, float Skew, float Strength,
                    bool bl_AplUniValue, bool bl_AreaPrView, float Gamma)
                    //bl_AplUniValue: apply uniform value?
                    //bl_AreaPrView: show target area?
                    //Gamma: perceptural correction or not?
{
    float p_rgb[3] = {r, g, b};
    float rgb[3];
    float Skew1;
    float Skew2;
    float adj_L;
    float targetV;
    float comparV;
    float amount;
    float Gamma2 = 1 / Gamma;
    Skew1 = Skew + 0.5;
    Skew2 = 1.0 - Skew1;
    adj_L = adj_rgb[3] + 1.0;
    
    if (param_choice == 0) { // Target: Red
        targetV = r + Pivot;
        comparV = g * Skew1 + b * Skew2;
    }else if  (param_choice == 1){ // Target: Green
        targetV = g + Pivot;
        comparV = r * Skew1 + b * Skew2;
    }else if  (param_choice == 2){ // Target: Blue
        targetV = b + Pivot;
        comparV = r * Skew1 + g * Skew2;
    }else if  (param_choice == 3){ // Target: Cyan
        comparV = r - Pivot;
        targetV = g * Skew2 + b * Skew1;
    }else if  (param_choice == 4){ // Target: Magenta
        comparV = g - Pivot;
        targetV = r * Skew2 + b * Skew1;
    }else if  (param_choice == 5){ // Target: Yellow
        comparV = b - Pivot;
        targetV = r * Skew2 + g * Skew1;
    }
    if (gamma22(targetV) > gamma22(comparV)){
        if (bl_AplUniValue == true){// apply uniform value
            amount = Strength;
        }else{
            amount = pow(targetV, Gamma2) - pow(comparV, Gamma2);
        }

        rgb = rgb_in_area(p_rgb, adj_rgb, adj_L, 
                        amount, Strength, bl_AreaPrView);
    }else  {
        rgb = rgb_out_area(p_rgb, bl_AreaPrView);
    }

    return rgb;
}

// @ART-param: ["param_choice", "対象色領域", ["赤", "緑", "青", "シアン", "マゼンタ", "黄"]]
// @ART-param: ["Pivot", "狭める <--> 拡げる", -0.25, 0.25, 0.0, 0.001, "対象色の閾値調整"]
// @ART-param: ["Skew", "", -0.5, 0.5, 0.0, 0.001, "対象色の傾き","target color: left --------- right\n    red: green sensitive  blue sensitive\n   green: red sensitive  blue sensitive\n    blue: red sensitive  green sensitive\n    cyan: green sensitive  blue sensitive\n magenta: red sensitive  blue sensitive\n yellow: red sensitive  green sensitive"]
// @ART-param: ["bl_AreaPrView", "対象領域表示", false]
// @ART-param: ["Gamma", "", 0.0, 2.2, 1.0, 0.001, "ガンマ補正"]
// @ART-param: ["R", "Red (シアン <--> 赤)", -1.0, 1.0, 0.0, 0.001, "対象色領域のRGB補正"]
// @ART-param: ["G", "Green (マゼンタ <--> 緑)", -1.0, 1.0, 0.0, 0.001, "対象色領域のRGB補正"]
// @ART-param: ["B", "Blue  (黄 <--> 青)", -1.0, 1.0, 0.0, 0.001, "対象色領域のRGB補正"]
// @ART-param: ["L", "明るさ  (暗 <--> 明)", -1.0, 1.0, 0.0, 0.001, "対象色領域のRGB補正"]
// @ART-param: ["Strength", "", 0.0, 3.0, 1.0, 0.001, "補正強度"]
// @ART-param: ["bl_AplUniValue", "均一の強度適用", false]


void ART_main(varying float r, varying float g, varying float b,
              int param_choice,
              float Pivot, float Skew, 
              bool bl_AreaPrView,
              float Gamma,
              float R, float G, float B, float L, 
              float Strength,
              bool bl_AplUniValue,
              output varying float rout,
              output varying float gout,
              output varying float bout)
{
    float rgb[3];
    float adj_rgb[4] = {R, G, B, L};


    rgb = colorCastCr(r, g, b, adj_rgb, param_choice, Pivot, Skew,
            Strength, bl_AplUniValue, bl_AreaPrView, Gamma);

    rout = rgb[0];
    gout = rgb[1];
    bout = rgb[2];
}
